@@ -1,11 +1,11 @@
 {- git-annex test suite framework
  -
- - Copyright 2010-2023 Joey Hess <id@joeyh.name>
+ - Copyright 2010-2024 Joey Hess <id@joeyh.name>
  -
  - Licensed under the GNU AGPL version 3 or higher.
  -}
 
-{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE OverloadedStrings, CPP #-}
 
 module Test.Framework where
 
@@ -67,6 +67,9 @@ import qualified Utility.Metered
 import qualified Utility.HumanTime
 import qualified Command.Uninit
 import qualified Utility.OsString as OS
+#ifndef mingw32_HOST_OS
+import qualified Utility.Gpg
+#endif
 
 -- Run a process. The output and stderr is captured, and is only
 -- displayed if the process does not return the expected value.
@@ -196,8 +199,10 @@ withtmpclonerepo' cfg a = do
 	case r of
 		Right () -> return ()
 		Left e -> do
-			whenM (keepFailuresOption . testOptions <$> getTestMode) $
-				putStrLn $ "** Preserving repo for failure analysis in " ++ clone
+			whenM (keepFailuresOption . testOptions <$> getTestMode) $ do
+				topdir <- Utility.Env.getEnvDefault "TOPDIR" ""
+				putStrLn $ "** Preserving repo for failure analysis in " ++ 
+					fromOsPath (toOsPath topdir </> toOsPath clone)
 			throwM e
 
 disconnectOrigin :: Assertion
@@ -227,9 +232,6 @@ inpath path a = do
 		Right v -> return v
 		Left e -> throwM e
 
-adjustedbranchsupported :: FilePath -> IO Bool
-adjustedbranchsupported repo = intopdir repo $ Annex.AdjustedBranch.isGitVersionSupported
-
 setuprepo :: FilePath -> IO FilePath
 setuprepo dir = do
 	cleanup dir
@@ -518,6 +520,33 @@ add_annex f faildesc = ifM (unlockedFiles <$> getTestMode)
 	, git_annex "add" [f] faildesc
 	)
 
+#ifndef mingw32_HOST_OS
+test_with_gpg :: (Utility.Gpg.GpgCmd -> [(String, String)] -> Assertion) -> Assertion
+test_with_gpg a = Utility.Tmp.Dir.withTmpDir (literalOsPath "gpgtmp") $ \gpgtmp -> do
+	-- Use the system temp directory as gpg temp directory because 
+	-- it needs to be able to store the agent socket there,
+	-- which can be problematic when testing some filesystems.
+	absgpgtmp <- absPath gpgtmp
+	res <- go absgpgtmp
+	-- gpg may still be running and would prevent
+	-- removeDirectoryRecursive from succeeding, so
+	-- force removal of the temp directory.
+	liftIO $ removeDirectoryForCleanup (fromOsPath gpgtmp)
+	return res
+  where
+	gpgcmd = Utility.Gpg.mkGpgCmd Nothing
+	go absgpgtmp = do
+		-- Since gpg uses a unix socket, which is limited to a
+		-- short path, use whichever is shorter of absolute
+		-- or relative path.
+		relgpgtmp <- relPathCwdToFile absgpgtmp
+		let gpgtmp = if OS.length relgpgtmp < OS.length absgpgtmp
+			then relgpgtmp 
+			else absgpgtmp
+		void $ Utility.Gpg.testHarness (fromOsPath gpgtmp) gpgcmd $ \environ ->
+			a gpgcmd environ
+#endif
+
 data TestMode = TestMode
 	{ unlockedFiles :: Bool
 	, adjustedUnlockedBranch :: Bool
@@ -536,6 +565,9 @@ testMode opts v = TestMode
 hasUnlockedFiles :: TestMode -> Bool
 hasUnlockedFiles m = unlockedFiles m || adjustedUnlockedBranch m
 
+onAdjustedBranch :: IO Bool
+onAdjustedBranch = adjustedUnlockedBranch <$> getTestMode
+
 withTestMode :: TestMode -> TestTree -> TestTree
 withTestMode testmode = withResource prepare release . const
   where
@@ -744,7 +776,7 @@ runFakeSsh ps = error $ "fake ssh option parse error: " ++ show ps
  - leave open are closed before finalCleanup is run at the end. This
  - prevents some failures to clean up after the test suite.
  -}
-parallelTestRunner :: TestOptions -> (Int -> Bool -> Bool -> TestOptions -> [TestTree]) -> IO ()
+parallelTestRunner :: TestOptions -> (Int -> Bool -> TestOptions -> [TestTree]) -> IO ()
 parallelTestRunner opts mkts = do
 	numjobs <- case concurrentJobs opts of
 		Just NonConcurrent -> pure 1
@@ -753,7 +785,7 @@ parallelTestRunner opts mkts = do
 		Nothing -> getNumProcessors
 	parallelTestRunner' numjobs opts mkts
 
-parallelTestRunner' :: Int -> TestOptions -> (Int -> Bool -> Bool -> TestOptions -> [TestTree]) -> IO ()
+parallelTestRunner' :: Int -> TestOptions -> (Int -> Bool -> TestOptions -> [TestTree]) -> IO ()
 parallelTestRunner' numjobs opts mkts
 	| fakeSsh opts = runFakeSsh (internalData opts)
 	| otherwise = go =<< Utility.Env.getEnv subenv
@@ -808,8 +840,7 @@ parallelTestRunner' numjobs opts mkts
 			<$> Annex.Init.probeCrippledFileSystem'
 				(toOsPath tmpdir)
 				Nothing Nothing False
-		adjustedbranchok <- Annex.AdjustedBranch.isGitVersionSupported
-		let ts = mkts numparts crippledfilesystem adjustedbranchok opts
+		let ts = mkts numparts crippledfilesystem opts
 		let warnings = fst (tastyParser ts)
 		unless (null warnings) $ do
 			hPutStrLn stderr "warnings from tasty:"
@@ -825,7 +856,7 @@ parallelTestRunner' numjobs opts mkts
 			let subdir = fromOsPath $ toOsPath tmpdir </> toOsPath (show n)
 			ensuredir subdir
 			let p = (proc pp ps)
-				{ env = Just ((subenv, show (n, crippledfilesystem, adjustedbranchok)):environ)
+				{ env = Just ((subenv, show (n, crippledfilesystem)):environ)
 				, cwd = Just subdir
 				}
 			(_, _, _, pid) <- createProcessConcurrent p
@@ -837,8 +868,8 @@ parallelTestRunner' numjobs opts mkts
 		return (length ts, exitcodes)
 	go (Just subenvval) = case readish subenvval of
 		Nothing -> error ("Bad " ++ subenv)
-		Just (n, crippledfilesystem, adjustedbranchok) -> setTestEnv $ do
-			let ts = mkts numparts crippledfilesystem adjustedbranchok opts
+		Just (n, crippledfilesystem) -> setTestEnv $ do
+			let ts = mkts numparts crippledfilesystem opts
 			let t = topLevelTestGroup [ ts !! (n - 1) ]
 			case tryIngredients ingredients tastyopts t of
 				Nothing -> error "No tests found!?"
